/*
 * ctcrelay.c
 *
 * Operate the relays on a US&S 506 Line Storage Unit
 * Created: 12/8/2011 2:33:08 PM
 *  Author: martin
 */ 

#include <avr/io.h>
//#define F_CPU 1000000UL // 1 MHz
//#define F_CPU 14.7456E6
#include <util/delay.h>

//relayon -- given relay number 0 -20, turn on the appropriate bit to energize the relay
void relayon(unsigned char relaynum) {
	// port D relay 0 - 7 (8 bits total)
	if(relaynum >= 0 && relaynum <=7) {
		PORTD |= (1<<relaynum);
	}
	else if(relaynum >=8 && relaynum <=13) {
		relaynum -= 8;
		PORTC |= (1 <<relaynum);
	}
	else if(relaynum >= 14 && relaynum <=20) {
		relaynum -= 14;
		PORTB |= (1 << relaynum);
	}
}

//relayoff -- given relay number 0 -20, turn off the appropriate bit to de-energize the relay
void relayoff(unsigned char relaynum) {
	// port D relay 0 - 7 (8 bits total)
	if(relaynum >= 0 && relaynum <=7) {
		PORTD &= ~(1<<relaynum);
	}
	else if(relaynum >=8 && relaynum <=13) {
		relaynum -= 8;
		PORTC &= ~(1 <<relaynum);
	}
	else if(relaynum >= 14 && relaynum <=20) {
		relaynum -= 14;
		PORTB &= ~(1 << relaynum);
	}
}
//relaytoggle -- given relay number 0 -20, toggle the appropriate bit to change state of the relay
void relaytoggle(unsigned char relaynum) {
	// port D relay 0 - 7 (8 bits total)
	if(relaynum >= 0 && relaynum <=7) {
		PORTD ^= (1<<relaynum);
	}
	else if(relaynum >=8 && relaynum <=13) {
		relaynum -= 8;
		PORTC ^= (1 <<relaynum);
	}
	else if(relaynum >= 14 && relaynum <=20) {
		relaynum -= 14;
		PORTB ^= (1 << relaynum);
	}
}

//dostep -- do a single step of the code line step 0 thru 15
// on step 15, all relays go off.
void dostep(unsigned char step) {
	unsigned char index;
	if(step < 16) {
		relaytoggle(0);	// toggle the line relay
		_delay_ms(50);
		relayon(1);
		_delay_ms(50);
		relayon(2);
		_delay_ms(40);
		relayon(3);
		_delay_ms(20);
		relayon(4);
		_delay_ms(90);
		relayoff(4);
		relayoff(3);
		if(step == 0 || (TCNT0 &1)) {
			// first step always has a long delay
			// after first step, randomly pick long or short step based on
			// low bit of timer0
			_delay_ms(150);
		}	
		
		// otherwise, short pulse, no delay
		
		relayon(step+5);		// turn on the step bit and leave it on
		_delay_ms(10);
		relayoff(2);
		_delay_ms(10);
		relayoff(1);
		_delay_ms(80);
	}
	else {	// step 16 turn off everything
		for(index = 20;index;index--) {
			relayoff(index);
			_delay_ms(10);
		}
		relayoff(0);
	}	
}
#define INPUTBIT (1<<PINB7)
#define IDLE (PINB & INPUTBIT)
#define CODE !(IDLE)
int main(void)
{
	DDRD=0xff;		//all of port D is output
	DDRC=0x3f;		//top two bits of port C don't exit
	DDRB=0x8f;		//top bit of port b is input
	PORTB = 0x80;	// enable pull up on port b7
	TCCR0B = CS00;	// Enable timer 0 running at 1mhz (we use the low bit of tcnt0 as a pseudo random on/off
    while(1)
    {
		unsigned char step;
		
		// wait until the code or indication line goes active
		if(CODE) {
			for(step = 0;step <= 8;step++) {
				dostep(step);	
			}
			dostep(16);
		}					
    }
}